/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_BLADE_DEBUG		FALSE

#include "grass/grass.pmh"

using namespace fe;

// NOTE dx and dv actually refer to angle and angular velocity,
// but we keep the familiar notation

Blade::Blade(void)
{
	//* current using uniform properties
	m_k=1e4;			// spring
	m_c=1e2;			// drag (angular)
	m_bend=0.1;			// added bend to rest angle
	m_m=1e0;			// mass
	m_h=1.0/60.0;		// time step
	m_t=0.0;			// time

	m_restoration=0.0;

	m_wind=0.0;
	m_turbulence=0.0;

	set(m_center);
	m_radius=0.0;

	populate();
}

void Blade::populate(void)
{
//	0	4	8	12
//	1	5	9	13
//	2	6	10	14
//	3	7	11	15

//	-2	1	0	0
//	1	-2	1	0   * k  (or c)
//	0	1	-2	1
//	0	0	1	-1

// non-symetrical?
//	-1	0	0	0
//	-1	-1	0	0   * restoration
//	-1	-1	-1	0
//	-1	-1	-1	-1

	set(m_dfdx);
	m_dfdx[0]= -2.0;
	m_dfdx[1]=1.0;

	m_dfdx[4]=1.0;
	m_dfdx[5]= -2.0;
	m_dfdx[6]=1.0;

	m_dfdx[9]=1.0;
	m_dfdx[10]= -2.0;
	m_dfdx[11]=1.0;

	m_dfdx[14]=1.0;
	m_dfdx[15]= -1.0;

	Matrix<4,4,Real> dfdv=m_dfdx;
	m_dfdx*=m_k;
	dfdv*=m_c;

	m_dfdx[0]-=m_restoration;
	m_dfdx[1]-=m_restoration;
	m_dfdx[2]-=m_restoration;
	m_dfdx[3]-=m_restoration;
	m_dfdx[5]-=m_restoration;
	m_dfdx[6]-=m_restoration;
	m_dfdx[7]-=m_restoration;
	m_dfdx[10]-=m_restoration;
	m_dfdx[11]-=m_restoration;
	m_dfdx[15]-=m_restoration;

	Matrix<4,4,Real> id;
	setIdentity(id);
	for(U32 m=0;m<4;m++)
	{
		id(m,m)*=4-m;
	}
	Matrix<4,4,Real> A=m_m*id-m_h*dfdv-m_h*m_h*m_dfdx;
	jad_invert(m_invA,A);

	Matrix<4,4,Real> check=A*m_invA;

#if FE_BLADE_DEBUG
	feLog("m_dfdx %s\n",c_print(m_dfdx));
	feLog("dfdv %s\n",c_print(dfdv));
	feLog("A %s\n",c_print(A));
	feLog("invA %s\n",c_print(m_invA));
	feLog("check %s\n",c_print(check));
#endif

	m_hdfdx=m_h*m_dfdx;
}

void Blade::computeAngles(U32 a_index, Vector<4,Real>& a_rAngle,
	const SpatialVector* a_pDelta)
{
	for(U32 m=0;m<4;m++)
	{
		// projected length onto ground
		const Real radialLength=sqrtf(a_pDelta[m][0]*a_pDelta[m][0]+
				a_pDelta[m][2]*a_pDelta[m][2]);

		SpatialVector projected(a_pDelta[m][0],0.0,a_pDelta[m][2]);
		const Real projectedLength=magnitude(projected);
		Real tandot=1.0;
		if(projectedLength>0.0)
		{
			tandot=dot(m_tangent[a_index],
					unit(SpatialVector(a_pDelta[m][0],0.0,a_pDelta[m][2])));
		}
		a_rAngle[m]=atan2f(tandot*radialLength,a_pDelta[m][1]);
	}
}

void Blade::storeAsRest(U32 a_index)
{
	//* TODO resize in larger grain
	if(m_rest.size()<a_index+1)
	{
		m_rest.resize(a_index+1);
	}
	if(m_length.size()<a_index+1)
	{
		m_length.resize(a_index+1);
	}
	if(m_contact.size()<a_index+1)
	{
		m_contact.resize(a_index+1);
	}

	//* TODO clean up duplication from step()

	const U32 index5=a_index*5;

	const SpatialVector& rLocation0=m_location[index5];
	const SpatialVector& rLocation1=m_location[index5+1];
	const SpatialVector& rLocation2=m_location[index5+2];
	const SpatialVector& rLocation3=m_location[index5+3];
	const SpatialVector& rLocation4=m_location[index5+4];

	// delta segment
	SpatialVector delta[4];
	delta[0]=rLocation1-rLocation0;
	delta[1]=rLocation2-rLocation1;
	delta[2]=rLocation3-rLocation2;
	delta[3]=rLocation4-rLocation3;

	Vector<4,Real> angle;
	computeAngles(a_index,angle,delta);

	//* rest angles are relative
	Vector<4,Real>& rRest=m_rest[a_index];
	rRest[0]=angle[0];
	rRest[1]=angle[1]-angle[0];
	rRest[2]=angle[2]-angle[1];
	rRest[3]=angle[3]-angle[2];

	//* segment length
	Vector<4,Real>& rLength=m_length[a_index];
	rLength[0]=sqrtf(delta[0][0]*delta[0][0]+
			delta[0][1]*delta[0][1]+delta[0][2]*delta[0][2]);
	rLength[1]=sqrtf(delta[1][0]*delta[1][0]+
			delta[1][1]*delta[1][1]+delta[1][2]*delta[1][2]);
	rLength[2]=sqrtf(delta[2][0]*delta[2][0]+
			delta[2][1]*delta[2][1]+delta[2][2]*delta[2][2]);
	rLength[3]=sqrtf(delta[3][0]*delta[3][0]+
			delta[3][1]*delta[3][1]+delta[3][2]*delta[3][2]);
}

void Blade::step(U32 a_index)
{
	const U32 index4=a_index*4;
	const U32 index5=a_index*5;

//	set(m_center,30.0*sin(m_t),0.0,30.0*cos(m_t));

	SpatialVector from;
	Vector<4,Real> push;	// external force
	set(push);

	SpatialVector& rLocation0=m_location[index5];
	SpatialVector& rLocation1=m_location[index5+1];
	SpatialVector& rLocation2=m_location[index5+2];
	SpatialVector& rLocation3=m_location[index5+3];
	SpatialVector& rLocation4=m_location[index5+4];

#if FALSE
	from=rLocation0-m_center;
	Real dist2=magnitudeSquared(from);
	if(dist2<m_radius*m_radius)
	{
		const Real alignment=dot(m_tangent[a_index],unit(from));
		setAll(push,5e2*(m_radius-sqrt(dist2))*alignment);
	}
#endif

	// delta segment
	SpatialVector delta[4];
	delta[0]=rLocation1-rLocation0;
	delta[1]=rLocation2-rLocation1;
	delta[2]=rLocation3-rLocation2;
	delta[3]=rLocation4-rLocation3;

	// segment length
	const Vector<4,Real>& l=m_length[a_index];

	Vector<4,Real> a;		// angle
	computeAngles(a_index,a,delta);

	Vector<4,Real> y;		// adjustment
	set(y);

	if(m_wind>0.0 || m_turbulence>0.0)
	{
		const Real locX=rLocation0[0];
		const Real locZ=rLocation0[2];
		const Real wind=m_tangent[a_index][0]*
				(m_wind+m_turbulence*sin(locX+m_t)*sin(locZ+m_t));
		push+=Vector<4,Real>(wind,wind,wind,wind);
	}

	push*=Vector<4,Real>(cos(a[0]),cos(a[1]),cos(a[2]),cos(a[3]));

	Vector<4,Real> w;		// angular velocity
	w[0]=m_velocity[index4];
	w[1]=m_velocity[index4+1];
	w[2]=m_velocity[index4+2];
	w[3]=m_velocity[index4+3];

	Vector<4,Real> rest=m_rest[a_index]+m_bend*Vector4f(1.0,1.0,1.0,1.0);

	Vector<4,Real> f;		// force
	f[0]=-m_k*(a[0]-rest[0])-m_c*(w[0])+
			m_k*(a[1]-a[0]-rest[1])+m_c*(w[1]-w[0])-
			m_restoration*(a[0]-rest[0]);

	f[1]=-m_k*(a[1]-a[0]-rest[1])-m_c*(w[1]-w[0])+
			m_k*(a[2]-a[1]-rest[2])+m_c*(w[2]-w[1])-
			m_restoration*(a[1]-2.0*rest[1]);

	f[2]=-m_k*(a[2]-a[1]-rest[2])-m_c*(w[2]-w[1])+
			m_k*(a[3]-a[2]-rest[3])+m_c*(w[3]-w[2])-
			m_restoration*(a[2]-3.0*rest[2]);

	f[3]=-m_k*(a[3]-a[2]-rest[3])-m_c*(w[3]-w[2])-
			m_restoration*(a[3]-4.0*rest[3]);
	f+=push;

	Vector<4,Real> b=m_h*(f+m_hdfdx*w+m_dfdx*y);
	Vector<4,Real> dv=m_invA*b;

#if FE_BLADE_DEBUG
	feLog("tangent %s\n",c_print(m_tangent[a_index]));
	feLog("push %s\n",c_print(push));
	feLog("f %s\n",c_print(f));
	feLog("b %s\n",c_print(b));
	feLog("dv %s\n",c_print(dv));
	feLog("l %s\n",c_print(l));
	feLog("a %s before\n",c_print(a));
	feLog("w %s\n",c_print(w));
#endif

	w+=dv;
	a+=m_h*w+y;

	//* TODO work this out properly
	const Real partX=m_tangent[a_index][0];
	const Real partZ=m_tangent[a_index][2];

	BWORD relocated=FALSE;
	for(U32 pass=0;pass<1+relocated;pass++)
	{
		const Real lsa0=l[0]*sin(a[0]);
		rLocation1[0]=rLocation0[0]+lsa0*partX;
		rLocation1[1]=rLocation0[1]+l[0]*cosf(a[0]);
		rLocation1[2]=rLocation0[2]+lsa0*partZ;

		const Real lsa1=l[1]*sin(a[1]);
		rLocation2[0]=rLocation1[0]+lsa1*partX;
		rLocation2[1]=rLocation1[1]+l[1]*cosf(a[1]);
		rLocation2[2]=rLocation1[2]+lsa1*partZ;

		const Real lsa2=l[2]*sin(a[2]);
		rLocation3[0]=rLocation2[0]+lsa2*partX;
		rLocation3[1]=rLocation2[1]+l[2]*cosf(a[2]);
		rLocation3[2]=rLocation2[2]+lsa2*partZ;

		const Real lsa3=l[3]*sin(a[3]);
		rLocation4[0]=rLocation3[0]+lsa3*partX;
		rLocation4[1]=rLocation3[1]+l[3]*cosf(a[3]);
		rLocation4[2]=rLocation3[2]+lsa3*partZ;

		if(pass)
		{
			break;
		}

		Contact& rContact=m_contact[a_index];

		//* collision displacement
		for(U32 m=0;m<4;m++)
		{
			const SpatialVector& base=m_location[index5+m];

			const SpatialVector toCenter=m_center-base;
			const Real reach=m_radius+l[m];
			if(magnitudeSquared(toCenter)<reach*reach)
			{

				//* point line-segment distance
				const Real sam=sinf(a[m]);
				const SpatialVector span(partX*sam,cosf(a[m]),partZ*sam);
				Real along=dot(toCenter,span)/l[m];
				if(along<0.0f)
				{
					from=base-m_center;
				}
				else if(along>1.0f)
				{
					from=base+l[m]*span-m_center;
				}
				else
				{
					SpatialVector projection=base+along*l[m]*span;
					from=projection-m_center;
				}

				Real distance=magnitude(from);

#if FE_BLADE_DEBUG
				feLog("%d angle %.6G dist %.6G along %.6G\n",
						m,a[m],distance,along);
#endif

				if(distance>0.0f && distance<m_radius)
				{
					relocated=TRUE;

					SpatialVector offset=from*(m_radius-distance)/distance;

					delta[m]=m_location[index5+1+m]-m_location[index5+m];

					const SpatialVector d_mod=delta[m]+offset;
					const Real r_mod=sqrtf(d_mod[0]*d_mod[0]+d_mod[2]*d_mod[2]);
					const Real tandot=dot(m_tangent[a_index],
							unit(SpatialVector(d_mod[0],0.0,d_mod[2])));
					const Real a_mod=atan2f(tandot*r_mod,d_mod[1]);

					y[m]=a_mod-a[m];
					w[m]=0.0;

#if FE_BLADE_DEBUG
					feLog("  offset %s\n",c_print(offset));
					feLog("  delta[m] %s\n",c_print(delta[m]));
					feLog("  d_mod %s\n",c_print(d_mod));
					feLog("  radial[m] %.6G\n",radial[m]);
					feLog("  r_mod %.6G\n",r_mod);
					feLog("  angle %.6G y[m] %.6G\n",a_mod,y[m]);
#endif
				}
				if(m==3)
				{
					if(along<0.99)
					{
						from=rLocation4-m_center;
						distance=magnitude(from);
					}
					//* TODO margin
					if(distance>0.0f && distance<m_radius)
					{
						if(!rContact.active())
						{
							rContact.set(m_radius*unit(rLocation4-m_center));
#if FE_BLADE_DEBUG
							feLog("contact set %d %s\n",
									a_index,c_print(rContact.offset()));
#endif
/*
							feLog(" center %s\n",c_print(m_center));
							feLog(" at %s\n",c_print(rLocation4));
							feLog(" to %s\n",
									c_print(m_center+rContact.offset()));
							feLog(" vs %s\n",c_print(m_center+
									distance*unit(rLocation4-m_center)));
							feLog(" distance %.6G/%.6G along %.6G\n",
									distance,m_radius,along);

							feLog(" relocated %d\n",relocated);
							feLog(" y %s\n",c_print(y));
							feLog(" distance %.6G from %s\n",
									distance,c_print(from));
							from=rLocation4-m_center;
							distance=magnitude(from);
							feLog(" distance %.6G from %s\n",
									distance,c_print(from));
*/
						}
					}
					else if(rContact.active() &&
							distance>m_radius*1.1)	//* tweak
					{
#if FE_BLADE_DEBUG
						feLog("contact near %d %s\n",
								a_index,c_print(rContact.offset()));
#endif
						rContact.clear();
					}
				}
			}
			else if(m==3 && rContact.active())
			{
#if FE_BLADE_DEBUG
				feLog("contact clear %d %s\n",
						a_index,c_print(rContact.offset()));
#endif
				rContact.clear();
			}
			if(relocated)
			{
				const Real lsam=l[m]*sin(a[m]);
				m_location[index5+1+m]=m_location[index5+m]+
						SpatialVector(lsam*partX,l[m]*cosf(a[m]),lsam*partZ);
			}
		}
		a+=y;

#if FE_BLADE_DEBUG
		feLog("a %s after\n",c_print(a));
		feLog("w %s\n",c_print(w));
#endif

		//* tip-constrained crunching
		if(rContact.active())
		{
#if FALSE
			const SpatialVector radial=rLocation0-m_center;
			const SpatialVector target=m_center+m_radius*unit(radial);
#else
			const SpatialVector target=m_center+rContact.offset();
#endif
			const SpatialVector span=target-rLocation0;
			const Real rise=span[1];
			const Real tandot=dot(m_tangent[a_index],
					unit(SpatialVector(span[0],0.0,span[2])));
			const Real tangential=tandot*sqrtf(span[0]*span[0]+span[2]*span[2]);
			const Real rotation=atan2f(tangential,rise);

			const Real distance=sqrtf(rise*rise+tangential*tangential);
			const Real ratio=(l[0]>0.0)? 0.25*distance/l[0]: 1.0;
			const Real radians=(ratio<1.0)? acos(0.25*distance/l[0]): 0.0;

			Real rounding=0.8-0.4*radians;	//* tweak
			rounding=(rounding>0.0)? rounding: 0.0;

			a[0]= -(1.0+rounding)*radians;
			a[1]= -(1.0-rounding)*radians;
			a[2]= -a[1];
			a[3]= -a[0];

/*
			feLog("target %s\n",c_print(target));
			feLog(" span %s\n",c_print(span));
			feLog(" tandot %.6G tangential %.6G rotation %.6G\n",
					tandot,tangential,rotation);
			feLog(" distance %.6G ratio %.6G radians %.6G rounding %.6G\n",
					distance,ratio,radians,rounding);
			feLog(" l %s\n",c_print(l));
			feLog(" a %s\n",c_print(a));
*/

			a+=rotation*Vector4f(1.0,1.0,1.0,1.0);

			set(w);
		}
	}

#if FALSE
	if(relocated)
	{
		const SpatialVector radial=rLocation0-m_center;
		const SpatialVector target=m_center+m_radius*unit(radial);

		InvKineCCD::solve(m_location,index5,5,target);
	}
#endif

	m_velocity[index4]=w[0];
	m_velocity[index4+1]=w[1];
	m_velocity[index4+2]=w[2];
	m_velocity[index4+3]=w[3];
}
