/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_OBLD_DEBUG			FALSE

#include "grass/grass.pmh"

using namespace fe;

void BladeOp::initialize(void)
{
	catalog<Real>("Tension")=1e4;
	catalog<Real>("Tension","high")=1e5;

	catalog<Real>("Drag")=1e2;
	catalog<Real>("Drag","high")=1e3;

	catalog<Real>("Bend")=1e-2;
	catalog<Real>("Bend","high")=1e0;

	catalog<Real>("Mass")=1e0;
	catalog<Real>("Mass","high")=1e1;

	catalog<Real>("Wind")=0;
	catalog<Real>("Wind","high")=1e3;

	catalog<Real>("Turbulence")=1e2;
	catalog<Real>("Turbulence","high")=1e3;

	catalog<Real>("TimeStep")=1.0/60.0;
	catalog<Real>("TimeStep","high")=1e-1;

	catalog<Real>("Restoration")=0.0;
	catalog<Real>("Restoration","high")=1e5;

	catalog<SpatialVector>("Center")=SpatialVector(0.0,0.0,0.0);

	catalog<Real>("Radius")=1e0;
	catalog<Real>("Radius","high")=1e1;

	catalog<String>("Tangent Attribute")="tangent";

	catalog< sp<Component> >("Input Curves");

	catalog< sp<Component> >("Output Curves");
	catalog<String>("Output Curves","IO")="output";
	catalog<String>("Output Curves","copy")="Input Curves";
	catalog<bool>("Output Curves","recycle")=true;
}

void BladeOp::handle(Record& a_rSignal)
{
#if FE_OBLD_DEBUG
	feLog("BladeOp::handle\n");
#endif

	//* access output curves (copy of input curves)
	m_spOutput=surfaceOutput(a_rSignal);
	if(m_spOutput.isNull())
	{
		feLog("OperatorDodge::handle"
				" output surface not SurfaceAccessibleI\n");
		return;
	}

	String tangentName=catalog<String>("Tangent Attribute");

	m_blade.setTension(catalog<Real>("Tension"));
	m_blade.setRestoration(catalog<Real>("Restoration"));
	m_blade.setDrag(catalog<Real>("Drag"));
	m_blade.setBend(catalog<Real>("Bend"));
	m_blade.setMass(catalog<Real>("Mass"));
	m_blade.setWind(catalog<Real>("Wind"));
	m_blade.setTurbulence(catalog<Real>("Turbulence"));
	m_blade.setCenter(catalog<SpatialVector>("Center"));
	m_blade.setRadius(catalog<Real>("Radius"));

	const Real timeStep=catalog<Real>("TimeStep");
	m_blade.setTimeStep(timeStep);

	static Real t=0.0;	// HACK
	t+=timeStep;
	m_blade.setTime(t);

	m_blade.populate();

	sp<SurfaceAccessorI> spAccessorTangent=m_spOutput->accessor(
			SurfaceAccessibleI::e_primitive,tangentName);

	sp<SurfaceAccessorI> spAccessorPoint=m_spOutput->accessor(
			SurfaceAccessibleI::e_point,SurfaceAccessibleI::e_position);

	sp<SurfaceAccessorI> spAccessorPrimitive=m_spOutput->accessor(
			SurfaceAccessibleI::e_primitive,SurfaceAccessibleI::e_vertices);

	const U32 pointCount=spAccessorPoint->count();
	const U32 primitiveCount=spAccessorPrimitive->count();
	const U32 subCount=primitiveCount? pointCount/primitiveCount: 0;

	std::vector<SpatialVector>& rTangent=m_blade.tangent();
	std::vector<SpatialVector>& rLocation=m_blade.location();
	std::vector<Real>& rVelocity=m_blade.velocity();

	//* inferred reset
	if(pointCount!=rLocation.size())
	{
		rLocation.resize(pointCount);
		for(U32 m=0;m<pointCount;m++)
		{
			rLocation[m]=spAccessorPoint->spatialVector(m,0);
		}

		const U32 pointCountV=pointCount*(subCount-1)/subCount;
		rVelocity.resize(pointCountV);
		for(U32 m=0;m<pointCountV;m++)
		{
			rVelocity[m]=0.0;
		}

		const U32 pointCountT=pointCount/subCount;
		rTangent.resize(pointCountT);
		if(spAccessorTangent.isValid() &&
				spAccessorTangent->count() &&
				spAccessorTangent->subCount(0))
		{
			for(U32 m=0;m<pointCountT;m++)
			{
				rTangent[m]=spAccessorTangent->spatialVector(m,0);
			}
		}
		else
		{
			feLog("BladeOp::handle tangent attribute \"%s\" missing\n",
					tangentName.c_str());

			for(U32 m=0;m<pointCountT;m++)
			{
				const Real angle=sin(0.7*m)+sin(0.3*m);

				set(rTangent[m],cos(angle),0.0,sin(angle));
			}
		}
		for(U32 m=0;m<pointCountT;m++)
		{
			m_blade.storeAsRest(m);
		}
	}

	setFullRange(0,primitiveCount-1);
	OperatorThreaded::handle(a_rSignal);

#if FE_OBLD_DEBUG
	feLog("BladeOp::handle done\n");
#endif
}

void BladeOp::run(I32 a_id,Range a_range)
{
	sp<SurfaceAccessorI> spAccessorPoint=m_spOutput->accessor(
			SurfaceAccessibleI::e_point,SurfaceAccessibleI::e_position);
	sp<SurfaceAccessorI> spAccessorPrimitive=m_spOutput->accessor(
			SurfaceAccessibleI::e_primitive,SurfaceAccessibleI::e_vertices);
	const U32 pointCount=spAccessorPoint->count();
	const U32 primitiveCount=spAccessorPrimitive->count();
	const U32 subCount=primitiveCount? pointCount/primitiveCount: 0;

	const I32 start=a_range.m_start;
	const I32 end=a_range.m_end;

	std::vector<SpatialVector>& rLocation=m_blade.location();

#if FE_OBLD_DEBUG
	feLog("BladeOp::run running thread %d [%d,%d]\n",a_id,start,end);
#endif

	for(U32 index=start;index<=end;index++)
	{
		m_blade.step(index);

		const U32 startSub=index*subCount;
		const U32 endSub=startSub+subCount;
		for(U32 indexSub=startSub;indexSub<endSub;indexSub++)
		{
			spAccessorPoint->set(indexSub,0,rLocation[indexSub]);
		}
	}

#if FE_OBLD_DEBUG
	feLog("BladeOp::run done running thread %d\n",a_id);
#endif
}
